#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

_logger = logging.getLogger(__name__)


def extract_response_field(response, field_name):
    if response is None:
        return None

    try:
        response_json = response.json()
        if isinstance(response_json, dict):
            return response_json.get(field_name)
        else:
            _logger.debug('HTTP response is not a dict: %s', str(response_json))
            return None
    except ValueError as e:
        _logger.debug('Failed to parse HTTP response: %s', e)
        return None
