#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class AssertionExtensions(object):
    # pylint:disable=no-member

    def assert_int_greater_than(self, x, y):
        self.assertGreater(x, y)
        self.assertNotEqual(float, type(x))

    def assert_int_greater_or_equal(self, x, y):
        self.assertGreaterEqual(x, y)
        self.assertNotEqual(float, type(x))

    def assert_float_greater_than(self, x, y):
        self.assertGreater(x, y)
        self.assertEqual(float, type(x))
